from durand.object_dictionary import Variable, ObjectDictionary
from durand.datatypes import DatatypeEnum as DT

# dynamic

od = ObjectDictionary()

state_control = Variable(0x2000, 0, DT.UNSIGNED8, 'rw')
state_actual = Variable(0x2001, 0, DT.UNSIGNED8, 'ro')


def update_state_control(value):
    print(f'Set control to {value}')


def read_state_actual():
    from random import randint
    return randint(0, 10)


od.add_object(state_control)
od.add_update_callback(state_control, update_state_control)

od.add_object(state_actual)
od.set_read_callback(state_actual, read_state_actual)

od.write(state_control, 5)
print(od.read(state_actual))


from dataclasses import dataclass
from collections import defaultdict
from typing import Any, Dict, List, Tuple, Callable
import logging

from .datatypes import DatatypeEnum, struct_dict, is_numeric, is_float
from .callback_handler import CallbackHandler, FailMode


log = logging.getLogger(__name__)


@dataclass(frozen=True)
class Variable:
    index: int
    subindex: int
    datatype: DatatypeEnum
    access: str
    default: Any = None
    factor: float = 1
    minimum: float = None
    maximum: float = None

    def __post_init__(self):
        if not 0 <= self.index <= 0xFFFF:
            raise ValueError("Index has to UINT16")
        if not 0 <= self.subindex <= 255:
            raise ValueError("Subindex has to be UINT8")
        if self.datatype not in DatatypeEnum:
            raise ValueError("Unsupported datatype")
        if self.access not in ("rw", "ro", "wo", "const"):
            raise ValueError("Invalid access type")
        if not is_numeric(self.datatype) and (
            self.maximum is not None or self.minimum is not None
        ):
            raise ValueError(
                "Minimum and Maximum not available with datatype %rs" % self.datatype
            )

    @property
    def multiplexor(self) -> Tuple[int, int]:
        return (self.index, self.subindex)

    @property
    def writable(self):
        return self.access in ("wo", "rw")

    @property
    def readable(self):
        return self.access in ("ro", "rw", "const")

    @property
    def size(self) -> int:
        if is_numeric(self.datatype):
            return struct_dict[self.datatype].size

        return None  # no size available

    def on_read(self, od: "ObjectDictionary"):
        # it's called via @variable.on_read(od) wrapping a function
        def wrap(func):
            od.set_read_callback(self, func)

        return wrap

    def on_update(self, od: "ObjectDictionary"):
        # it's called via @variable.on_update(od) wrapping a function
        def wrap(func):
            od.update_callbacks[self].add(func)

        return wrap

    def pack(self, value) -> bytes:
        if not is_numeric(self.datatype):
            return bytes(value)

        value = value / self.factor
        dt_struct = struct_dict[self.datatype]

        if not is_float(self.datatype):
            value = int(value)

        return dt_struct.pack(value)

    def unpack(self, data: bytes):
        if not is_numeric(self.datatype):
            return bytes(data)

        dt_struct = struct_dict[self.datatype]
        value = dt_struct.unpack(data)[0]
        value *= self.factor

        return value


class ObjectDictionary:
    def __init__(self):
        self._variables: Dict[Tuple[int, int], Variable] = dict()
        self._data: Dict[Variable, Any] = dict()

        self.validate_callbacks: Dict[Variable, CallbackHandler] = defaultdict(
            lambda: CallbackHandler(fail_mode=FailMode.FIRST_FAIL)
        )
        self.update_callbacks: Dict[Variable, CallbackHandler] = defaultdict(
            CallbackHandler
        )
        self.download_callbacks: Dict[Variable, CallbackHandler] = defaultdict(
            CallbackHandler
        )
        self._read_callbacks: Dict[Variable, Callable] = dict()

    def add_object(self, variable: Variable):
        self._variables[variable.multiplexor] = variable
        if variable.default is not None:
            self._data[variable] = variable.default
        else:
            self._data[variable] = 0 if is_numeric(variable.datatype) else b""

        if variable.subindex == 0:
            return

        if (variable.index, 0) in self._variables:
            largest_subindex = self._variables[(variable.index, 0)]
        else:
            largest_subindex = Variable(
                index=variable.index,
                subindex=0,
                datatype=DatatypeEnum.UNSIGNED8,
                access="const",
            )

            self._variables[(variable.index, 0)] = largest_subindex

        value = max(self._data.get(largest_subindex, 0), variable.subindex)
        self._data[largest_subindex] = value

    def lookup(self, index: int, subindex: int = 0) -> Variable:
        return self._variables[(index, subindex)]

    def write(self, variable: Variable, value: Any, downloaded: bool = True):
        self.validate_callbacks[variable].call(value)  # may raises exception
        self._data[variable] = value
        self.update_callbacks[variable].call(value)

        if not downloaded:
            return

        self.download_callbacks[variable].call(value)

    def read(self, variable: Variable):
        if variable in self._read_callbacks:
            return self._read_callbacks[variable]()

        return self._data[variable]

    def set_read_callback(self, variable: Variable, callback) -> None:
        self._read_callbacks[variable] = callback

    @property
    def variables(self) -> Tuple[Variable]:
        return tuple(self._variables.keys())



from typing import List

from .adapters import AdapterABC
from .object_dictionary import ObjectDictionary
from .services.sdo import SDOServer
from .services.pdo import TPDO, RPDO
from .services.nmt import NMTService, StateEnum
from .services.heartbeat import HeartbeatProducer
from .object_dictionary import Variable
from .datatypes import DatatypeEnum as DT


SDO_SERVERS = 127

class Node:
    def __init__(self, adapter: AdapterABC, node_id: int, od: ObjectDictionary = None):

        self.adapter = adapter
        self.node_id = node_id
        od = ObjectDictionary() if od is None else od
        self.object_dictionary = od

        self.nmt = NMTService(self)
        self.tpdo = {i: TPDO(self, i) for i in range(1, 5)}
        self.rpdo = {i: RPDO(self, i) for i in range(1, 5)}

        self.sdo_servers: List[SDOServer] = list()

        self.sdo_servers.append(
            SDOServer(self, 0, 0x600 + self.node_id, 0x580 + self.node_id)
        )

        assert 1 <= SDO_SERVERS <= 127, "Number of SDO servers has to be between 1 and 127"

        for index in range(1, SDO_SERVERS + 1):
            self.sdo_servers.append(SDOServer(self, index))

        HeartbeatProducer(self)

        od.add_object(Variable(0x1000, 0, DT.UNSIGNED32, "ro", 0))  # device type
        od.add_object(Variable(0x1001, 0, DT.UNSIGNED8, "ro", 0))  # error register
        od.add_object(
            Variable(0x1018, 1, DT.UNSIGNED32, "ro", 0)
        )  # identity - vendor-id

        self.nmt.set_state(StateEnum.PRE_OPERATIONAL)



from struct import Struct
from enum import IntEnum


class DatatypeEnum(IntEnum):
    BOOLEAN = 1
    INTEGER8 = 2
    INTEGER16 = 3
    INTEGER32 = 4
    INTEGER64 = 0x15
    UNSIGNED8 = 5
    UNSIGNED16 = 6
    UNSIGNED32 = 7
    UNSIGNED64 = 0x1B
    VISIBLE_STRING = 9
    OCTET_STRING = 10
    REAL32 = 8
    REAL64 = 0x11
    DOMAIN = 15


def is_numeric(datatype: DatatypeEnum):
    return datatype not in (
        DatatypeEnum.VISIBLE_STRING,
        DatatypeEnum.OCTET_STRING,
        DatatypeEnum.DOMAIN,
    )


def is_float(datatype: DatatypeEnum):
    return datatype in (DatatypeEnum.REAL32, DatatypeEnum.REAL64)


struct_dict = {
    DatatypeEnum.UNSIGNED8: Struct("B"),
    DatatypeEnum.INTEGER8: Struct("b"),
    DatatypeEnum.UNSIGNED16: Struct("<H"),
    DatatypeEnum.INTEGER16: Struct("<h"),
    DatatypeEnum.UNSIGNED32: Struct("<I"),
    DatatypeEnum.INTEGER32: Struct("<i"),
    DatatypeEnum.UNSIGNED64: Struct("<Q"),
    DatatypeEnum.INTEGER64: Struct("<q"),
    DatatypeEnum.REAL32: Struct("<f"),
    DatatypeEnum.REAL64: Struct("<d"),
    DatatypeEnum.VISIBLE_STRING: Struct("s"),
    DatatypeEnum.OCTET_STRING: Struct("s"),
    DatatypeEnum.DOMAIN: Struct("s"),
}



from typing import Dict, Callable
from unittest.mock import Mock

from durand.adapters import AdapterABC


class MockAdapter(AdapterABC):
    def __init__(self):
        self.subscriptions = dict()
        self.tx_mock = Mock()

    def add_subscription(self, cob_id: int, callback):
        self.subscriptions[cob_id] = callback

    def remove_subscription(self, cob_id: int):
        self.subscriptions.pop(cob_id)

    def receive(self, cob_id: int, msg: bytes):
        """Used in tests to send a CAN message to the node"""
        self.subscriptions[cob_id](cob_id, msg)

    def send(self, cob_id: int, msg: bytes):
        """When the node is sending a CAN message, it will be captured in .tx_mock"""
        self.tx_mock(cob_id, msg)



from unittest.mock import Mock, call
import struct
from binascii import crc_hqx

import pytest

from durand import Node, Variable
from durand.datatypes import DatatypeEnum as DT
from durand.datatypes import struct_dict
from durand.services.sdo.server import SDO_STRUCT
from durand.services.sdo import BaseDownloadHandler, BaseUploadHandler

from .adapter import MockAdapter


def build_sdo_packet(
    cs: int, index: int, subindex: int = 0, data: bytes = b"", toggle: bool = False
):
    cmd = (cs << 5) + (toggle << 4)

    if data:
        cmd += ((4 - len(data)) << 2) + 3  # set size and expetited

    return SDO_STRUCT.pack(cmd, index, subindex) + data + bytes(4 - len(data))


@pytest.mark.parametrize("node_id", [0x01, 0x7F])
def test_sdo_object_dictionary(node_id):
    n = Node(MockAdapter(), node_id)

    assert n.object_dictionary.lookup(0x1200, 0) == Variable(
        0x1200, 0, DT.UNSIGNED8, "const", default=2
    )
    assert n.object_dictionary.lookup(0x1200, 1) == Variable(
        0x1200, 1, DT.UNSIGNED32, "ro", default=0x600 + node_id
    )
    assert n.object_dictionary.lookup(0x1200, 2) == Variable(
        0x1200, 2, DT.UNSIGNED32, "ro", default=0x580 + node_id
    )


@pytest.mark.parametrize("node_id", [0x01, 0x7F])
@pytest.mark.parametrize("index", [1, 2, 127])
def test_sdo_additional_servers(node_id, index):
    n = Node(MockAdapter(), node_id)

    assert n.object_dictionary.lookup(0x1200 + index, 0) == Variable(
        0x1200 + index, 0, DT.UNSIGNED8, "const", default=3
    )
    assert n.object_dictionary.lookup(0x1200 + index, 1) == Variable(
        0x1200 + index, 1, DT.UNSIGNED32, "rw", default=0x8000_0000
    )
    assert n.object_dictionary.lookup(0x1200 + index, 2) == Variable(
        0x1200 + index, 2, DT.UNSIGNED32, "rw", default=0x8000_0000
    )
    assert n.object_dictionary.lookup(0x1200 + index, 3) == Variable(
        0x1200 + index, 3, DT.UNSIGNED8, "rw"
    )


@pytest.mark.parametrize(
    "datatype",
    [
        DT.UNSIGNED8,
        DT.INTEGER8,
        DT.UNSIGNED16,
        DT.INTEGER16,
        DT.UNSIGNED32,
        DT.INTEGER32,
        DT.REAL32,
    ],
)
@pytest.mark.parametrize("with_handler", [True, False])
def test_sdo_download_expetited(datatype, with_handler):
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    handler_mock = Mock()
    handler_calls = list()

    v = Variable(0x2000, 0, datatype, "rw")
    n.object_dictionary.add_object(v)

    def download_handler(node, variable, size):
        assert node == n
        assert variable == v
        return handler_mock

    if with_handler:
        n.sdo_servers[0].download_manager.set_handler_callback(download_handler)

    mock_write = Mock()
    n.object_dictionary.update_callbacks[v].add(mock_write)

    mock_write.assert_not_called()
    adapter.tx_mock.reset_mock()

    # with specified size
    size = struct_dict[datatype].size
    cmd = 0x23 + ((4 - size) << 2)
    adapter.receive(0x602, cmd.to_bytes(1, "little") + b"\x00\x20\x00\x01\x02\x03\x04")

    handler_calls.append(call.on_receive(b"\x01\x02\x03\x04"[:size]))
    handler_calls.append(call.on_finish())

    adapter.tx_mock.assert_called_once_with(0x582, b"\x60\x00\x20\x00\x00\x00\x00\x00")
    value = struct_dict[datatype].unpack(b"\x01\x02\x03\x04"[:size])[0]

    if with_handler:
        mock_write.assert_not_called()
        assert handler_mock.mock_calls == handler_calls
    else:
        mock_write.assert_called_once_with(value)
        assert not handler_mock.mock_calls

    # without specified size
    adapter.tx_mock.reset_mock()
    mock_write.reset_mock()
    handler_mock.reset_mock()

    adapter.receive(0x602, b"\x22\x00\x20\x00\x04\x03\x02\x01")
    adapter.tx_mock.assert_called_once_with(0x582, b"\x60\x00\x20\x00\x00\x00\x00\x00")

    handler_calls = [call.on_receive(b"\x04\x03\x02\x01"[:size]), call.on_finish()]

    value = struct_dict[datatype].unpack(b"\x04\x03\x02\x01"[:size])[0]

    if with_handler:
        mock_write.assert_not_called()
        assert handler_mock.mock_calls == handler_calls
    else:
        mock_write.assert_called_once_with(value)
        assert not handler_mock.mock_calls


def test_sdo_download_fails():
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    # invalid client command specifier
    adapter.receive(0x602, build_sdo_packet(cs=7, index=0x1200, data=b"AB"))
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x12\x00\x01\x00\x04\x05")

    # write 'const' and 'ro' variables
    adapter.receive(
        0x602, build_sdo_packet(cs=1, index=0x1200, data=b"AB")
    )  # const entry
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x12\x00\x02\x00\x01\x06")

    adapter.receive(
        0x602, build_sdo_packet(cs=1, index=0x1200, subindex=1, data=b"AB")
    )  # ro entry
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x12\x01\x02\x00\x01\x06")

    # data size not matching
    var = Variable(0x2000, 0, DT.UNSIGNED8, "rw", minimum=0x10, maximum=0x20)
    n.object_dictionary.add_object(var)
    adapter.receive(
        0x602, build_sdo_packet(cs=1, index=0x2000, data=b"AB")
    )  # write too many bytes to unsigned8
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x20\x00\x10\x00\x07\x06")

    # value to low or high
    adapter.receive(
        0x602, build_sdo_packet(cs=1, index=0x2000, data=b"\x0F")
    )  # write 15 (below minimum of 16)
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x20\x00\x32\x00\x09\x06")

    adapter.receive(
        0x602, build_sdo_packet(cs=1, index=0x2000, data=b"\x21")
    )  # write 33 (above maximum of 32)
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x20\x00\x31\x00\x09\x06")

    adapter.tx_mock.reset_mock()

    adapter.receive(
        0x602, build_sdo_packet(cs=1, index=0x2000, data=b"\x10")
    )  # write 16 (ok with minimum of 16)
    adapter.receive(
        0x602, build_sdo_packet(cs=1, index=0x2000, data=b"\x20")
    )  # write 32 (ok with maximum of 32)
    adapter.tx_mock.assert_has_calls(
        [
            call(0x582, build_sdo_packet(3, 0x2000)),
            call(0x582, build_sdo_packet(3, 0x2000)),
        ]
    )

    # write not working
    def fail(value):
        raise ValueError("Validation failed")

    n.object_dictionary.validate_callbacks[var].add(fail)
    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2000, data=b"\x10"))
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x20\x00\x20\x00\x00\x08")

    # object not existing
    adapter.receive(
        0x602, build_sdo_packet(cs=1, index=0x2001, data=b"\x10")
    )  # write non existing object
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x01\x20\x00\x00\x00\x02\x06")

    # subindex not existing
    adapter.receive(
        0x602, build_sdo_packet(cs=1, index=0x2000, subindex=1, data=b"\x10")
    )  # write non existing object
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x20\x01\x11\x00\x09\x06")


@pytest.mark.parametrize("datatype", [DT.VISIBLE_STRING, DT.DOMAIN, DT.OCTET_STRING])
def test_sdo_download_segmented(datatype):
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    var = Variable(0x2000, 0, datatype, "rw")
    n.object_dictionary.add_object(var)

    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2000, subindex=0))
    adapter.tx_mock.assert_called_with(0x582, b"\x60\x00\x20\x00\x00\x00\x00\x00")

    adapter.receive(0x602, b"\x00ABCDEFG")
    adapter.tx_mock.assert_called_with(0x582, b"\x20\x00\x00\x00\x00\x00\x00\x00")

    adapter.receive(0x602, b"\x1AHI")
    adapter.tx_mock.assert_called_with(0x582, b"\x30\x00\x00\x00\x00\x00\x00\x00")

    assert n.object_dictionary.read(var) == b""

    adapter.receive(0x602, b"\x07JKLM")
    adapter.tx_mock.assert_called_with(0x582, b"\x20\x00\x00\x00\x00\x00\x00\x00")

    assert n.object_dictionary.read(var) == b"ABCDEFGHIJKLM"


@pytest.mark.parametrize(
    "datatype",
    [
        DT.UNSIGNED8,
        DT.INTEGER8,
        DT.UNSIGNED16,
        DT.INTEGER16,
        DT.UNSIGNED32,
        DT.INTEGER32,
        DT.REAL32,
    ],
)
def test_sdo_download_segmented_numeric(datatype):
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    v = Variable(0x2000, 0, datatype, "rw")
    n.object_dictionary.add_object(v)

    mock_write = Mock()
    n.object_dictionary.update_callbacks[v].add(mock_write)

    mock_write.assert_not_called()
    adapter.tx_mock.reset_mock()

    # with specified size
    size = struct_dict[datatype].size
    adapter.receive(0x602, b"\x21\x00\x20\x00" + struct.pack("<I", size))
    adapter.tx_mock.assert_called_once_with(0x582, b"\x60\x00\x20\x00\x00\x00\x00\x00")

    value = struct_dict[datatype].unpack(b"\x01\x02\x03\x04"[:size])[0]

    for index in range(size):
        cmd = 0x0C + ((index % 2) << 4) + (index == size - 1)
        adapter.receive(
            0x602,
            cmd.to_bytes(1, "little")
            + b"\x01\x02\x03\x04"[index : index + 1]
            + bytes(6),
        )

        response = 0x20 + ((index % 2) << 4)
        adapter.tx_mock.assert_called_with(
            0x582, response.to_bytes(1, "little") + bytes(7)
        )

    mock_write.assert_called_once_with(value)

    # without specified size
    adapter.tx_mock.reset_mock()
    mock_write.reset_mock()

    adapter.receive(0x602, b"\x20\x00\x20\x00\x00\x00\x00\x00")
    adapter.tx_mock.assert_called_once_with(0x582, b"\x60\x00\x20\x00\x00\x00\x00\x00")

    value = struct_dict[datatype].unpack(b"\x04\x03\x02\x01"[:size])[0]

    for index in range(size):
        cmd = 0x0C + ((index % 2) << 4) + (index == size - 1)
        adapter.receive(
            0x602,
            cmd.to_bytes(1, "little")
            + b"\x04\x03\x02\x01"[index : index + 1]
            + bytes(6),
        )

        response = 0x20 + ((index % 2) << 4)
        adapter.tx_mock.assert_called_with(
            0x582, response.to_bytes(1, "little") + bytes(7)
        )

    mock_write.assert_called_once_with(value)


def test_sdo_download_segmented_fails():
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    # write 'const' and 'ro' variables
    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x1200))  # const entry
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x12\x00\x02\x00\x01\x06")

    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x1200, subindex=1))  # ro entry
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x12\x01\x02\x00\x01\x06")

    # data size not matching
    var1 = Variable(0x2000, 0, DT.UNSIGNED8, "rw", minimum=0x10, maximum=0x20)
    n.object_dictionary.add_object(var1)

    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2000))
    adapter.receive(0x602, b"\x01" + bytes(7))  # write too many bytes to unsigned8
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x20\x00\x10\x00\x07\x06")

    # value to low or high
    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2000))
    adapter.receive(0x602, b"\x0D\x0E" + bytes(6))  # write 15 (below minimum of 16)
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x20\x00\x32\x00\x09\x06")

    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2000))
    adapter.receive(0x602, b"\x0D\x21" + bytes(6))  # write 33 (above maximum of 32)
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x20\x00\x31\x00\x09\x06")

    adapter.tx_mock.reset_mock()

    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2000))
    adapter.receive(0x602, b"\x0D\x10" + bytes(6))  # write 16 (ok with minimum of 16)

    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2000))
    adapter.receive(0x602, b"\x0D\x20" + bytes(6))  # write 32 (ok with maximum of 32)

    adapter.tx_mock.assert_has_calls(
        [
            call(0x582, build_sdo_packet(3, 0x2000)),
            call(0x582, b"\x20" + bytes(7)),
            call(0x582, build_sdo_packet(3, 0x2000)),
            call(0x582, b"\x20" + bytes(7)),
        ]
    )

    # download segment without init
    adapter.receive(0x602, b"\x0D\x10" + bytes(6))  # write 16
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x00\x00\x01\x00\x04\x05")

    # wrong toggle bit (initial)
    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2000))
    adapter.receive(0x602, b"\x1D\x10" + bytes(6))  # write 16 with toggle bit set
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x20\x00\x00\x00\x03\x05")

    # write not working
    var2 = Variable(0x2001, 0, DT.UNSIGNED8, "rw")
    n.object_dictionary.add_object(var2)

    def fail(value):
        raise ValueError("Validation failed")

    n.object_dictionary.validate_callbacks[var2].add(fail)
    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2001))
    adapter.receive(0x602, b"\x0D\x10" + bytes(6))  # write 16
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x01\x20\x00\x20\x00\x00\x08")

    # abort
    adapter.tx_mock.reset_mock()

    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2000))
    adapter.receive(0x602, b"\x80\x00\x20\x00\x01\x02\x03\x04")
    adapter.receive(0x602, b"\x0D\x12" + bytes(6))  # write 18

    adapter.tx_mock.assert_has_calls(
        [
            call(0x582, build_sdo_packet(3, 0x2000)),
            call(0x582, b"\x80\x00\x00\x00\x01\x00\x04\x05"),
        ]
    )
    assert n.object_dictionary.read(var1) == 32

    # abort another transfer
    adapter.tx_mock.reset_mock()

    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2000))
    adapter.receive(0x602, b"\x80\x00\x20\x01\x01\x02\x03\x04")
    adapter.receive(0x602, b"\x0D\x11" + bytes(6))  # write 17

    adapter.tx_mock.assert_has_calls(
        [call(0x582, build_sdo_packet(3, 0x2000)), call(0x582, b"\x20" + bytes(7))]
    )
    assert n.object_dictionary.read(var1) == 17


def test_sdo_download_segmented_handler():
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    var1 = Variable(0x2000, 0, DT.DOMAIN, "rw")
    var2 = Variable(0x2001, 0, DT.DOMAIN, "rw")

    n.object_dictionary.add_object(var1)
    n.object_dictionary.add_object(var2)

    handler_mock = Mock()

    def download_callback(node: None, variable: Variable, size: int):
        if variable.index == 0x2001:
            return handler_mock

    n.sdo_servers[0].download_manager.set_handler_callback(download_callback)

    # test without Downloadhandler
    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2000))
    adapter.receive(0x602, b"\x0D\x10" + bytes(6))

    assert n.object_dictionary.read(var1) == b"\x10"

    handler_mock.assert_not_called()

    # test wit Downloadhandler
    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2001))
    adapter.receive(0x602, b"\x0D\x11" + bytes(6))

    assert n.object_dictionary.read(var2) == b""
    handler_mock.on_receive.assert_called_once_with(b"\x11")
    handler_mock.on_finish.assert_called_once_with()
    handler_mock.on_abort.assert_not_called()

    # aborting Downloadhandler
    handler_mock.reset_mock()

    adapter.receive(0x602, build_sdo_packet(cs=1, index=0x2001))
    adapter.receive(0x602, b"\x80\x01\x20\x00\x01\x02\x03\x04")
    adapter.receive(0x602, b"\x1D\x11" + bytes(6))
    adapter.receive(0x602, b"\x80\x01\x20\x00\x01\x02\x03\x04")

    assert n.object_dictionary.read(var2) == b""
    handler_mock.on_receive.assert_not_called()
    handler_mock.on_finish.assert_not_called()
    handler_mock.on_abort.assert_called_once_with()


@pytest.mark.parametrize(
    "size", [1, 7, 8, 889, 890, 889 * 2, 889 * 2 + 1, 4096, 10_000]
)
@pytest.mark.parametrize("crc", [True, False])
@pytest.mark.parametrize("with_handler", [True, False])
@pytest.mark.parametrize("with_size", [True, False])
def test_sdo_download_block(size, crc, with_handler, with_size):
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    var = Variable(0x2000, 0, DT.DOMAIN, "rw")
    n.object_dictionary.add_object(var)

    handler_mock = Mock()
    handler_calls = list()

    def download_handler(node, variable, size):
        assert node == n
        assert variable == var
        return handler_mock

    if with_handler:
        n.sdo_servers[0].download_manager.set_handler_callback(download_handler)

    adapter.tx_mock.reset_mock()

    cmd = 0xC0 + (crc << 2) + (with_size << 1)
    size_bytes = struct.pack("<I", size) if with_size else bytes(4)

    adapter.receive(0x602, cmd.to_bytes(1, "little") + b"\x00\x20\x00" + size_bytes)
    adapter.tx_mock.assert_called_once_with(0x582, b"\xA4\x00\x20\x00\x7F\x00\x00\x00")

    for _ in range((size - 1) // 889):
        adapter.tx_mock.reset_mock()

        for sub_block_index in range(1, 128):
            adapter.tx_mock.assert_not_called()
            adapter.receive(0x602, sub_block_index.to_bytes(1, "little") + b"\xAA" * 7)
            handler_calls.append(call.on_receive(b"\xaa" * 7))

        adapter.tx_mock.assert_called_once_with(
            0x582, b"\xA2\x7F\x7F\x00\x00\x00\x00\x00"
        )

    adapter.tx_mock.reset_mock()

    for sub_block_index in range(1, ((size - 1) % 889) // 7 + 1 if size else 0):
        adapter.tx_mock.assert_not_called()
        adapter.receive(0x602, sub_block_index.to_bytes(1, "little") + b"\xAA" * 7)
        handler_calls.append(call.on_receive(b"\xaa" * 7))

    adapter.tx_mock.reset_mock()

    if size != 0:
        segment_nr = ((size - 1) % 889) // 7 + 1
        cmd = 0x80 + segment_nr
        adapter.receive(
            0x602,
            cmd.to_bytes(1, "little")
            + b"\xAA" * ((size - 1) % 7 + 1)
            + bytes(6 - (size - 1) % 7),
        )
        handler_calls.append(call.on_receive(b"\xaa" * ((size - 1) % 7 + 1)))

        adapter.tx_mock.assert_called_once_with(
            0x582,
            b"\xA2" + segment_nr.to_bytes(1, "little") + b"\x7F\x00\x00\x00\x00\x00",
        )

    cmd = 0xC1 + ((6 - (size - 1) % 7) << 2)

    if crc and size:
        crc_bytes = struct.pack("<H", crc_hqx(b"\xAA" * size, 0))
    else:
        crc_bytes = bytes(2)

    adapter.tx_mock.reset_mock()
    adapter.receive(0x602, cmd.to_bytes(1, "little") + crc_bytes + bytes(5))
    handler_calls.append(call.on_finish())

    adapter.tx_mock.assert_called_once_with(0x582, b"\xA1" + bytes(7))

    if with_handler:
        assert n.object_dictionary.read(var) == b""
        assert handler_mock.mock_calls == handler_calls
    else:
        assert n.object_dictionary.read(var) == b"\xAA" * size


def test_sdo_block_failed():
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    # write 'const' and 'ro' variables
    adapter.receive(0x602, build_sdo_packet(cs=6, index=0x1200))  # const entry
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x12\x00\x02\x00\x01\x06")

    adapter.receive(0x602, build_sdo_packet(cs=6, index=0x1200, subindex=1))  # ro entry
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x12\x01\x02\x00\x01\x06")

    # wrong sequence number
    var = Variable(0x2000, 0, DT.DOMAIN, "rw")
    n.object_dictionary.add_object(var)

    adapter.tx_mock.reset_mock()

    adapter.receive(0x602, b"\xC0\x00\x20\x00" + bytes(4))
    adapter.tx_mock.assert_called_once_with(0x582, b"\xa4\x00\x20\x00\x7f\x00\x00\x00")
    adapter.tx_mock.reset_mock()
    adapter.receive(0x602, b"\x00" + bytes(7))
    adapter.tx_mock.assert_called_once_with(0x582, b"\x80\x00\x20\x00\x03\x00\x04\x05")

    # block end with init
    adapter.tx_mock.reset_mock()

    adapter.receive(0x602, b"\xC1" + bytes(7))
    adapter.tx_mock.assert_called_once_with(0x582, b"\x80\x00\x00\x00\x01\x00\x04\x05")

    # wrong crc
    adapter.receive(0x602, b"\xC4\x00\x20\x00" + bytes(4))
    adapter.receive(0x602, b"\x81\xBB" + bytes(6))
    adapter.tx_mock.reset_mock()
    adapter.receive(0x602, b"\xD9" + bytes(7))
    adapter.tx_mock.assert_called_once_with(0x582, b"\x80\x00\x20\x00\x04\x00\x04\x05")

    # failed on_receive (1)
    adapter.tx_mock.reset_mock()

    handler_mock = Mock()
    handler_mock.on_receive.side_effect = ValueError()

    def download_handler(node, variable, size):
        return handler_mock

    n.sdo_servers[0].download_manager.set_handler_callback(download_handler)

    adapter.receive(0x602, b"\xC4\x00\x20\x00" + bytes(4))
    adapter.receive(0x602, b"\x01\xBB" + bytes(6))
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x20\x00\x20\x00\x00\x08")


@pytest.mark.parametrize(
    "datatype",
    [
        DT.UNSIGNED8,
        DT.INTEGER8,
        DT.UNSIGNED16,
        DT.INTEGER16,
        DT.UNSIGNED32,
        DT.INTEGER32,
        DT.REAL32,
    ],
)
def test_sdo_upload_expetited(datatype):
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    value = struct_dict[datatype].unpack(
        b"\x01\x02\x03\x04"[: struct_dict[datatype].size]
    )[0]

    v = Variable(0x2000, 0, datatype, "rw", default=value)
    n.object_dictionary.add_object(v)

    mock_write = Mock()
    n.object_dictionary.update_callbacks[v].add(mock_write)

    mock_write.assert_not_called()
    adapter.tx_mock.reset_mock()

    # with specified size
    size = struct_dict[datatype].size
    adapter.receive(0x602, b"\x40\x00\x20\x00\x00\x00\x00\x00")

    cmd = 0x43 + ((4 - size) << 2)
    adapter.tx_mock.assert_called_once_with(
        0x582,
        cmd.to_bytes(1, "little")
        + b"\x00\x20\x00"
        + b"\x01\x02\x03\x04"[:size]
        + bytes(4 - size),
    )


def test_sdo_upload_fails():
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    # read 'wo' variables
    v = Variable(0x2000, 0, DT.INTEGER8, "wo", default=0)
    n.object_dictionary.add_object(v)

    adapter.receive(0x602, build_sdo_packet(cs=2, index=0x2000))  # wo entry
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x00\x20\x00\x01\x00\x01\x06")

    # read not working
    v = Variable(0x2001, 0, DT.INTEGER8, "rw", default=0)
    n.object_dictionary.add_object(v)

    def fail():
        raise ValueError("Validation failed")

    n.object_dictionary.set_read_callback(v, fail)
    adapter.receive(0x602, build_sdo_packet(cs=2, index=0x2001))
    adapter.tx_mock.assert_called_with(0x582, b"\x80\x01\x20\x00\x20\x00\x00\x08")


def test_upload_segmented():
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    v = Variable(0x2000, 0, DT.DOMAIN, "ro", default=b"ABCDEFGHIJKLMNO")
    n.object_dictionary.add_object(v)

    adapter.receive(0x602, build_sdo_packet(cs=2, index=0x2000))  # init upload
    adapter.tx_mock.assert_called_with(
        0x582, b"\x41\x00\x20\x00\x0f\x00\x00\x00"
    )  # response to init upload

    adapter.receive(0x602, b"\x60\x00\x20\x00\x00\x00\x00\x00")  # first segment
    adapter.tx_mock.assert_called_with(
        0x582, b"\x00ABCDEFG"
    )  # response for first segment

    adapter.receive(0x602, b"\x70\x00\x20\x00\x00\x00\x00\x00")  # second segment
    adapter.tx_mock.assert_called_with(
        0x582, b"\x10HIJKLMN"
    )  # response for second segment

    adapter.receive(0x602, b"\x60\x00\x20\x00\x00\x00\x00\x00")  # third segment
    adapter.tx_mock.assert_called_with(
        0x582, b"\x0DO\x00\x00\x00\x00\x00\x00"
    )  # response for third segment


class UploadHandler(BaseUploadHandler):
    def __init__(self, data):
        self.data = data
        self.size = len(data)

    def on_read(self, size: int):
        response_data, self.data = self.data[:size], self.data[size:]
        return response_data


@pytest.mark.parametrize("with_handler", [True, False])
@pytest.mark.parametrize("size", [1, 2, 6, 7, 8, 14, 100])
@pytest.mark.parametrize("with_pst", [True, False])  # protocol switching threshold
def test_upload_segmented_various_length(with_handler, size, with_pst):
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    v = Variable(0x2000, 0, DT.DOMAIN, "ro", default=b"")
    n.object_dictionary.add_object(v)

    def handler_callback(node: Node, variable: Variable):
        if variable == v:
            return UploadHandler(b"\xAA" * size)

    if with_handler:
        n.sdo_servers[0].upload_manager.set_handler_callback(handler_callback)
    else:
        n.object_dictionary.write(v, b"\xAA" * size, False)

    if with_pst:
        adapter.receive(
            0x602, struct.pack("<BHBBBH", 0xA0, 0x2000, 0, 127, size, 0)
        )  # init block upload with protocol switching threshold
    else:
        adapter.receive(
            0x602, build_sdo_packet(cs=2, index=0x2000)
        )  # init segmented upload

    if size <= 4:
        cmd = 0x43 + ((4 - size) << 2)
        data = b"\xAA" * size + bytes(4 - size)
        adapter.tx_mock.assert_called_with(
            0x582, cmd.to_bytes(1, "little") + b"\x00\x20\x00" + data
        )  # response to init upload (expitited)
        return

    adapter.tx_mock.assert_called_with(
        0x582, b"\x41\x00\x20\x00" + struct.pack("<I", size)
    )  # response to init upload

    toggle = False

    for _ in range((size - 1) // 7):
        cmd = 0x60 + (toggle << 4)

        adapter.receive(
            0x602, cmd.to_bytes(1, "little") + b"\x00\x20\x00\x00\x00\x00\x00"
        )  # request a segment
        adapter.tx_mock.assert_called_with(
            0x582, (toggle << 4).to_bytes(1, "little") + b"\xAA" * 7
        )  # response segment
        toggle = not toggle

    cmd = 0x60 + (toggle << 4)
    adapter.receive(
        0x602, cmd.to_bytes(1, "little") + b"\x00\x20\x00\x00\x00\x00\x00"
    )  # request last segment

    cmd = 0x01 + ((6 - ((size - 1) % 7)) << 1) + (toggle << 4)
    adapter.tx_mock.assert_called_with(
        0x582,
        cmd.to_bytes(1, "little")
        + b"\xAA" * ((size - 1) % 7 + 1)
        + b"\x00" * (6 - (size - 1) % 7),
    )  # response for last segment


@pytest.mark.parametrize("with_handler", [True, False])
def test_block_upload(with_handler):
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    v = Variable(0x2000, 0, DT.DOMAIN, "ro", default=b"ABCDEFGHIJKLMNO")
    n.object_dictionary.add_object(v)

    def handler_callback(node: Node, variable: Variable):
        if variable == v:
            return UploadHandler(b"ABCDEFGHIJKLMNO")

    if with_handler:
        n.sdo_servers[0].upload_manager.set_handler_callback(handler_callback)

    adapter.receive(
        0x602, b"\xA4\x00\x20\x00\x02\x0D\x00\x00"
    )  # request block with crc, blocksize=2, pst=14
    adapter.tx_mock.assert_called_with(
        0x582, b"\xC6\x00\x20\x00\x0f\x00\x00\x00"
    )  # response to block upload init

    adapter.tx_mock.reset_mock()
    adapter.receive(0x602, b"\xA3" + bytes(7))  # request first part of block
    calls = [call(0x582, b"\x01ABCDEFG"), call(0x582, b"\x02HIJKLMN")]
    assert adapter.tx_mock.mock_calls == calls

    adapter.receive(
        0x602, b"\xA2\x02\x02" + bytes(5)
    )  # acknowledge fist part of blocks
    adapter.tx_mock.assert_called_with(
        0x582, b"\x81O" + bytes(6)
    )  # response for second part of blocks

    adapter.receive(
        0x602, b"\xA2\x01\x02" + bytes(5)
    )  # acknowledge second part of blocks
    crc = crc_hqx(b"ABCDEFGHIJKLMNO", 0)
    adapter.tx_mock.assert_called_with(
        0x582, b"\xD9" + crc.to_bytes(2, "little") + b"\x00\x00\x00\x00\x00"
    )  # response for third segment

    adapter.tx_mock.reset_mock()
    adapter.receive(0x602, b"\xA1" + bytes(7))  # acknowled transfer
    adapter.tx_mock.assert_not_called()


@pytest.mark.parametrize("size", [0, 1, 2, 6, 7, 8, 14, 100])
@pytest.mark.parametrize("blocksize", [1, 127])
@pytest.mark.parametrize("with_crc", [True, False])
@pytest.mark.parametrize(
    "with_handler, with_size", [(True, True), (True, False), (False, True)]
)
def test_block_upload_extended(size, blocksize, with_crc, with_handler, with_size):
    adapter = MockAdapter()
    n = Node(adapter, 0x02)

    data = (b"ABC" * (size // 3 + 1))[:size]
    crc = crc_hqx(data, 0) if with_crc else 0

    v = Variable(0x2000, 0, DT.DOMAIN, "ro", default=data)
    n.object_dictionary.add_object(v)

    def handler_callback(node: Node, variable: Variable):
        if variable == v:
            handler = UploadHandler(data)
            if not with_size:
                handler.size = None
            return handler

    if with_handler:
        n.sdo_servers[0].upload_manager.set_handler_callback(handler_callback)

    cmd = 0xA0 + (with_crc << 2)
    pst = max(min(size - 1, 255), 0)
    adapter.receive(
        0x602, struct.pack("<BHBBBH", cmd, 0x2000, 0, blocksize, pst, 0)
    )  # request block with/without crc, various blocksize, pst=size-1

    size_response = size if with_size else 0
    cmd_response = 0xC4 + (with_size << 1)
    adapter.tx_mock.assert_called_with(
        0x582,
        cmd_response.to_bytes(1, "little")
        + b"\x00\x20\x00"
        + size_response.to_bytes(4, "little"),
    )  # response to block upload init

    adapter.tx_mock.reset_mock()
    adapter.receive(0x602, b"\xA3" + bytes(7))  # request upload block

    if not data:
        adapter.tx_mock.assert_called_with(
            0x582, (0x81).to_bytes(1, "little") + bytes(7)
        )
        adapter.receive(
            0x602, b"\xA2" + struct.pack("<BB", blocksize, blocksize) + bytes(5)
        )  # request next upload block

    while data:
        block_data, data = data[: blocksize * 7], data[blocksize * 7 :]
        slices = [block_data[i : i + 7] for i in range(0, len(block_data), 7)]

        calls = []

        for index, slice in enumerate(slices[:-1]):
            calls.append(call(0x582, (index + 1).to_bytes(1, "little") + slice))

        calls.append(
            call(
                0x582,
                (((not data) << 7) + len(slices)).to_bytes(1, "little")
                + slices[-1]
                + bytes(7 - len(slices[-1])),
            )
        )

        assert adapter.tx_mock.mock_calls == calls

        adapter.tx_mock.reset_mock()
        adapter.receive(
            0x602, b"\xA2" + struct.pack("<BB", blocksize, blocksize) + bytes(5)
        )  # request next upload block

    cmd = 0xC1 + ((7 - (size % 7)) << 2)
    adapter.tx_mock.assert_called_with(
        0x582, struct.pack("<BH", cmd, crc) + bytes(5)
    )  # response to last block upload requst



from abc import ABCMeta, abstractmethod
from typing import Dict, Callable


class AdapterABC(metaclass=ABCMeta):
    @abstractmethod
    def add_subscription(self, cob_id: int, callback):
        """add subscription
        :param cob_id: cob_id to subscribe
        :param callback: to be called when cob_id is received
        """

    @abstractmethod
    def remove_subscription(self, cob_id: int):
        """remove subscription
        :param cob_id: remove subscription for cob_id
        """

    @abstractmethod
    def send(self, cob_id: int, msg: bytes):
        """sending a CAN message to the adapter
        :param cob_id: CAN arbitration id
        :param msg: CAN data bytes
        """



from enum import Enum
import logging


log = logging.getLogger(__name__)


class FailMode(Enum):
    IGNORE = 1  # exceptions are ignored
    FIRST_FAIL = 2  # first callback raising a exception will stop calling the following and escalate the exception
    LATE_FAIL = (
        3  # all callbacks will be called, but the first exception will be escalated
    )


class CallbackHandler:
    def __init__(self, fail_mode: FailMode = FailMode.IGNORE):
        self._callbacks = list()
        self._fail_mode = fail_mode

    def add(self, callback):
        self._callbacks.append(callback)

    def remove(self, callback):
        self._callbacks.remove(callback)

    def call(self, *args, **kwargs):
        exception = None

        for callback in self._callbacks:
            try:
                callback(*args, **kwargs)
            except Exception as e:
                if self._fail_mode == FailMode.LATE_FAIL and exception is None:
                    exception = e
                elif self._fail_mode == FailMode.FIRST_FAIL:
                    raise e
                else:
                    log.debug("Ignored exception in callback handler", exc_info=True)

        if exception:
            raise exception



from typing import TYPE_CHECKING

from durand.object_dictionary import Variable
from durand.datatypes import DatatypeEnum as DT
from durand.scheduler import get_scheduler

if TYPE_CHECKING:
    from durand.node import Node


class HeartbeatProducer:
    def __init__(self, node: "Node"):
        self._handle = None
        self._node = node

        heartbeat_producer = Variable(0x1017, 0, DT.UNSIGNED16, "rw")

        node.object_dictionary.add_object(heartbeat_producer)
        node.object_dictionary.update_callbacks[heartbeat_producer].add(
            self._update_interval
        )

    def _update_interval(self, value: int):
        sched = get_scheduler()

        if self._handle:
            sched.cancel(self._handle)
            self._handle = None

        if value:
            self._process_heartbeat(value / 1000)

    def _process_heartbeat(self, interval: float):
        msg = bytes((self._node.nmt.state,))  # data contains NMT state
        self._node.adapter.send(0x700 + self._node.node_id, msg)

        self._handle = get_scheduler().add(
            interval, self._process_heartbeat, args=(interval,)
        )



from enum import IntEnum
import logging
from typing import TYPE_CHECKING

from ..callback_handler import CallbackHandler

if TYPE_CHECKING:
    from ..node import Node


log = logging.getLogger(__name__)


class StateEnum(IntEnum):
    INITIALISATION = 0
    STOPPED = 4
    OPERATIONAL = 5
    PRE_OPERATIONAL = 127


class NMTService:
    def __init__(self, node: "Node"):
        self._node = node
        self.state_callbacks = CallbackHandler()

        self.state = None

        node.adapter.add_subscription(cob_id=0, callback=self.handle_msg)

        self.set_state(StateEnum.INITIALISATION)

    def handle_msg(self, cob_id: int, msg: bytes):
        cs, node_id = msg[:2]

        if node_id not in (0, self._node.node_id):  # 0 is used for broadcast
            return

        if cs == 0x01:  # start node
            self.set_state(StateEnum.OPERATIONAL)
        elif cs == 0x02:  # stop node
            self.set_state(StateEnum.STOPPED)
        elif cs == 0x80:  # enter pre-operational
            self.set_state(StateEnum.PRE_OPERATIONAL)
        elif cs in (0x81, 0x82):  # Reset Node or Reset Communication
            self.set_state(StateEnum.INITIALISATION)
            self.set_state(StateEnum.PRE_OPERATIONAL)
        else:
            log.error("Unknown NMT command specifier 0x%02X", cs)

    def set_state(self, state: StateEnum):
        if state == self.state:
            return

        if state == StateEnum.INITIALISATION:
            # send bootup message
            self._node.adapter.send(0x700 + self._node.node_id, b"\x00")

        self.state_callbacks.call(state)

        self.state = state



from dataclasses import dataclass
from re import match
from datetime import datetime

from durand import Node, Variable, DatatypeEnum


def datetime_to_time(d: datetime):
    return d.strftime("%I:%M") + ("AM" if d.hour < 12 else "PM")


def datetime_to_date(d: datetime):
    return d.strftime("%m-%d-%Y")


@dataclass
class FileInfo:
    FileName: str = "python_durand_device.eds"
    FileVersion: int = 0
    FileRevision: int = 0
    EDSVersion: str = "4.0"
    Description: str = None
    CreationTime: str = None
    CreationDate: str = None
    CreatedBy: str = None
    ModificationTime: str = None
    ModificationDate: str = None
    ModifiedBy: str = None

    Comment: str = None

    def validate(self):
        if not 0 <= self.FileVersion <= 255:
            raise ValueError("FileVersion is Unsigned8")
        if not 0 <= self.FileRevision <= 255:
            raise ValueError("FileRevision is Unsigned8")
        if not match("[d].[d]", self.EDSVersion):
            raise ValueError("EDSVersion type mismatch")

        if self.CreationDate:
            try:
                datetime.strptime("%m-%d-%Y", self.CreationDate)
            except ValueError:
                raise ValueError("CreationDate format invalid")

        if self.CreationTime:
            if len(self.CreationTime) != 7 or self.CreationTime[5:] not in ("AM", "PM"):
                raise ValueError("CreationTime format invalid")

            try:
                datetime.strptime("%I:%M", self.CreationTime[:6])
            except ValueError:
                raise ValueError("CreationTime format invalid")

        if self.ModificationDate:
            try:
                datetime.strptime("%m-%d-%Y", self.ModificationDate)
            except ValueError:
                raise ValueError("ModificationDate format invalid")

        if self.ModificationTime:
            if len(self.ModificationTime) != 7 or self.ModificationTime[5:] not in (
                "AM",
                "PM",
            ):
                raise ValueError("ModificationTime format invalid")

            try:
                datetime.strptime("%I:%M", self.ModificationTime[:6])
            except ValueError:
                raise ValueError("ModificationTime format invalid")


class EDS:
    def __init__(self):
        self.FileInfo = FileInfo()


class EDSProvider:
    def __init__(self, node: Node, eds: EDS = None):
        self._node = node
        self.eds = eds or EDS()

        eds_store_variable = Variable(0x1021, 0, DatatypeEnum.DOMAIN, "ro")
        self._node.object_dictionary.add_object(eds_store_variable)
        self._node.object_dictionary.set_read_callback(
            eds_store_variable, self.generate
        )



from cgitb import handler
import struct
from binascii import crc_hqx
from typing import Tuple

from durand.datatypes import is_numeric
from durand.object_dictionary import Variable

from .server import SDODomainAbort, SDOServer, TransferState, SDO_STRUCT


class BaseUploadHandler:
    size = None

    def on_read(self, size: int) -> bytes:
        """on_read is called when new data is requested

        :param size: number of bytes to be read
        :returns: next slice of data
        """

    def on_abort(self) -> None:
        """on_abort is called when the transfer was aborted"""


class HandlerStream:
    def __init__(self, upload_handler: BaseUploadHandler):
        self._handler = upload_handler
        self.size = upload_handler.size
        self._buffer = bytearray()

    def _extend(self, size: int):
        missing_bytes = size - len(self._buffer)

        if missing_bytes > 0:
            self._buffer.extend(self._handler.on_read(missing_bytes))

    def peek(self, size: int) -> bytes:
        self._extend(size)
        return self._buffer[:size]

    def read(self, size: int) -> bytes:
        self._extend(size)
        data = self._buffer[:size]
        del self._buffer[:size]
        return data

    def abort(self):
        self._handler.on_abort()
        self._buffer.clear()

    def release(self):
        self._buffer.clear()


class FixedStream:
    def __init__(self, data: bytes):
        self._buffer = memoryview(data)
        self.size = len(self._buffer)

    def peek(self, size: int) -> bytes:
        return bytes(self._buffer[:size])

    def read(self, size: int) -> bytes:
        data = bytes(self._buffer[:size])
        self._buffer = self._buffer[size:]
        return data

    def abort(self):
        self._buffer.release()

    def release(self):
        self._buffer.release()


class UploadManager:
    def __init__(self, server: SDOServer):
        self._server = server

        self._handler_callback = None

        self._variable = None
        self._state = TransferState.NONE

        # used for block transfer
        self._block_size = None
        self._crc = None

        # used for segmented transfer
        self._toggle_bit = None

    def set_handler_callback(self, callback):
        self._handler_callback = callback

    @property
    def block_transfer_active(self):
        return self._state in (TransferState.BLOCK, TransferState.BLOCK_END)

    def on_abort(self, variable):
        if self._state != TransferState.NONE and self._variable == variable:
            self._abort()

    def _abort(self):
        self._stream.abort()
        self._stream = None
        self._state = TransferState.NONE

    def setup(self, index, subindex) -> int:
        variable = self._server.lookup(index, subindex)

        if variable.access == "wo":
            raise SDODomainAbort(0x06010001)  # read a write-only object

        self._variable = variable

        if self._handler_callback:
            handler = self._handler_callback(self._server.node, variable)

            if handler:
                self._stream = HandlerStream(handler)
                return

        try:
            value = self._server.node.object_dictionary.read(variable)
        except:
            raise SDODomainAbort(0x08000020, variable)  # data can't be transferred

        if is_numeric(variable.datatype):
            value = variable.pack(value)

        self._stream = FixedStream(value)

    def init_upload(self, msg: bytes):
        if self._state != TransferState.NONE:
            self._abort()

        cmd, index, subindex = SDO_STRUCT.unpack(msg[:4])

        self.setup(index, subindex)
        size = self._stream.size

        if msg[0] & 0xE3 == 0xA0 and (
            msg[5] == 0 or size is None or size > msg[5]
        ):  # ccs=5, cs=0 -> block upload
            # msg[5] is protocol switching threshold
            self._state = TransferState.BLOCK
            self._crc = 0 if msg[0] & 0x04 else None

            size_bytes = struct.pack("<I", size) if size is not None else bytes(4)

            self._block_size = msg[4]

            cmd = 0xC4 if size is None else 0xC6
            self._server.node.adapter.send(
                self._server.cob_tx, cmd.to_bytes(1, "little") + msg[1:4] + size_bytes
            )
            return

        if size is not None and size <= 4:
            # make an expetited transfer
            data = self._stream.read(4)
            self._stream.release()
            self._stream = None
            cmd = 0x43 + ((4 - len(data)) << 2)
            response = (
                SDO_STRUCT.pack(cmd, index, subindex) + data + bytes(4 - len(data))
            )
            self._server.node.adapter.send(self._server.cob_tx, response)
            return

        self._state = TransferState.SEGMENT
        self._toggle_bit = False

        cmd = 0x40 + (size is not None)
        size_bytes = struct.pack("<I", size) if size is not None else bytes(4)

        self._server.node.adapter.send(
            self._server.cob_tx, cmd.to_bytes(1, "little") + msg[1:4] + size_bytes
        )

    def upload_segment(self, msg: bytes):
        if self._state != TransferState.SEGMENT:
            self._abort()
            raise SDODomainAbort(
                0x05040001, variable=False
            )  # client command specificer not valid

        toggle_bit = bool(msg[0] & 0x10)

        if toggle_bit != self._toggle_bit:
            self._abort()
            raise SDODomainAbort(0x05030000, self._variable)  # toggle bit not altered

        self._toggle_bit = not self._toggle_bit

        data = self._stream.read(7)
        no_data_left = not self._stream.peek(1)

        if no_data_left:
            self._stream.release()
            self._stream = None

        cmd = (toggle_bit << 4) + no_data_left + ((7 - len(data)) << 1)

        self._server.node.adapter.send(
            self._server.cob_tx, cmd.to_bytes(1, "little") + data + bytes(7 - len(data))
        )

    def upload_sub_block(self, msg: bytes):
        if self._state == TransferState.BLOCK_END:
            self._stream.release()
            self._stream = None
            self._state = TransferState.NONE
            return

        if self._state != TransferState.BLOCK:
            self._abort()
            raise SDODomainAbort(
                0x05040001, variable=None
            )  # client command specificer not valid

        if msg[0] & 0x03 == 2:
            data = self._stream.read(msg[1] * 7)
            if self._crc is not None:
                self._crc = crc_hqx(data, self._crc)

            self._block_size = msg[2]

            if not self._stream.peek(1):
                cmd = 0xC1 + ((7 - len(data) % 7) << 2)
                crc = self._crc if self._crc is not None else 0
                self._server.node.adapter.send(
                    self._server.cob_tx,
                    cmd.to_bytes(1, "little") + crc.to_bytes(2, "little") + bytes(5),
                )
                self._state = TransferState.BLOCK_END
                return

        data = self._stream.peek(self._block_size * 7 + 1)
        data, following_byte = (
            data[: self._block_size * 7],
            data[self._block_size * 7 :],
        )
        chunks = [data[i : i + 7] for i in range(0, len(data), 7)]

        for index, chunk in enumerate(chunks[:-1]):
            response = (index + 1).to_bytes(1, "little") + chunk
            self._server.node.adapter.send(self._server.cob_tx, response)

        last_block = not following_byte
        first_byte = (last_block << 7) + (len(chunks) if chunks else 1)
        data = chunks[-1] if chunks else b""
        response = first_byte.to_bytes(1, "little") + data + bytes(7 - len(data))
        self._server.node.adapter.send(self._server.cob_tx, response)



import asyncio
from abc import ABCMeta, abstractmethod, abstractproperty
from sched import scheduler
from typing import TypeVar
import functools
import threading


TEntry = TypeVar("TEntry")  # type of scheduler entry


class AbstractScheduler(metaclass=ABCMeta):
    @abstractmethod
    def add(self, delay: float, callback, args=(), kwargs={}) -> TEntry:
        """Add a new scheduler entry.

        :param delay: time in seconds when the callback should be called
        :param callback: a function object called when time has passed
        :param args: tuple with positional arguments for the callback
        :param kwargs: dictionary with keyword arguments
        :returns: an id which can be used to cancel the scheduled callback
        """

    @abstractmethod
    def cancel(self, entry: TEntry):
        """Cancel a scheduled callback

        :param entry: the scheduler entry to be canceld
        """

    def start(self):
        """Start scheduling"""

    @abstractproperty
    def lock(self):
        """A global lock which can be used the assure thread safety"""


class AsyncScheduler(AbstractScheduler):
    def __init__(self, loop=None):
        self._loop = loop

        self._lock = threading.Lock()

    def add(self, delay: float, callback, args=(), kwargs=None) -> asyncio.TimerHandle:
        if kwargs:
            callback = functools.partial(callback, **kwargs)

        loop = self._loop or asyncio.get_event_loop()
        return loop.call_later(delay, callback, *args)

    def cancel(self, entry: asyncio.TimerHandle):
        entry.cancel()

    @property
    def lock(self):
        return self._lock


_scheduler = AsyncScheduler()


def get_scheduler() -> AbstractScheduler:
    return _scheduler


def set_scheduler(scheduler: AbstractScheduler):
    global _scheduler
    _scheduler = scheduler



import struct
from binascii import crc_hqx

from durand.object_dictionary import Variable
from .server import SDODomainAbort, SDOServer, TransferState, SDO_STRUCT


class BaseDownloadHandler:
    def on_receive(self, data: bytes):
        """on_received is called when new data is received

        :param data: bytes received and to be appended
        """

    def on_finish(self):
        """on_finish is called when the transfer is successfully completed"""

    def on_abort(self):
        """on_abort is called when the transfer was aborted"""


class DownloadManager:
    def __init__(self, server: SDOServer):
        self._server = server

        self._handler_callback = None
        self._handler: BaseDownloadHandler = None

        self._variable = None
        self._state = TransferState.NONE

        self._buffer = bytearray()

        # used for block transfer
        self._sequence_number = None
        self._crc = None

        # used for segmented transfer
        self._toggle_bit = None

    def set_handler_callback(self, callback):
        self._handler_callback = callback

    @property
    def block_transfer_active(self):
        return self._state == TransferState.BLOCK

    def _init(self, new_state: TransferState):
        self._state = new_state

        self._handler = None
        self._sequence_number = None
        self._toggle_bit = False
        self._buffer.clear()

    def on_abort(self, variable):
        if self._state != TransferState.NONE and self._variable == variable:
            self._abort()

    def _receive(self, data: bytes):
        if self._handler:
            try:
                self._handler.on_receive(data)
            except:
                self._abort()
                raise SDODomainAbort(0x08000020, self._variable)  # data can't be stored
        else:
            self._buffer.extend(data)

    def _abort(self):
        if self._handler:
            self._handler.on_abort()

        self._init(TransferState.NONE)

    def _finish(self):
        try:
            if self._handler:
                self._handler.on_finish()
            else:
                value = self._variable.unpack(self._buffer)

                if (
                    self._variable.minimum is not None
                    and value < self._variable.minimum
                ):
                    raise SDODomainAbort(0x06090032, self._variable)  # value too low

                if (
                    self._variable.maximum is not None
                    and value > self._variable.maximum
                ):
                    raise SDODomainAbort(0x06090031, self._variable)  # value too high

                self._server.node.object_dictionary.write(self._variable, value)
        except struct.error:
            raise SDODomainAbort(
                0x06070010, self._variable
            )  # datatype size is not matching
        except SDODomainAbort as e:
            raise e
        except Exception as e:
            raise SDODomainAbort(0x08000020, self._variable)  # data can't be stored
        finally:
            self._init(TransferState.NONE)

    def init_download(self, msg: bytes):
        if self._state != TransferState.NONE:
            self._abort()

        cmd, index, subindex = SDO_STRUCT.unpack(msg[:4])

        variable = self._server.lookup(index, subindex)

        if variable.access not in ("rw", "wo"):
            raise SDODomainAbort(0x06010002)  # write a read-only object

        response = SDO_STRUCT.pack(0x60, index, subindex) + bytes(4)

        if not cmd & 0x02:  # segmented transfer (not expitited)
            self._init(TransferState.SEGMENT)
            self._variable = variable

            if cmd & 0x01:  # size specified
                size = int.from_bytes(msg[4:], "little")
            else:
                size = None

            if self._handler_callback:
                self._handler = self._handler_callback(
                    self._server.node, variable, size
                )

            self._server.node.adapter.send(self._server.cob_tx, response)
            return

        if cmd & 0x01:  # size specified
            size = 4 - ((cmd >> 2) & 0x03)
        else:
            size = variable.size

        self._init(TransferState.NONE)

        if self._handler_callback:
            self._handler = self._handler_callback(self._server.node, variable, size)

        self._variable = variable
        self._receive(msg[4 : 4 + size])
        self._finish()

        self._server.node.adapter.send(self._server.cob_tx, response)

    def download_segment(self, msg: bytes):
        if self._state != TransferState.SEGMENT:
            self._abort()
            raise SDODomainAbort(
                0x05040001, variable=False
            )  # client command specificer not valid

        toggle_bit = bool(msg[0] & 0x10)

        if toggle_bit != self._toggle_bit:
            self._abort()
            raise SDODomainAbort(0x05030000, self._variable)  # toggle bit not altered

        self._toggle_bit = not self._toggle_bit

        size = 7 - ((msg[0] & 0x0E) >> 1)

        self._receive(msg[1 : 1 + size])

        if msg[0] & 0x01:  # check continue bit
            self._finish()

        cmd = 0x20 + (toggle_bit << 4)
        self._server.node.adapter.send(
            self._server.cob_tx, cmd.to_bytes(1, "little") + bytes(7)
        )

    def download_block_init(self, msg: bytes):
        if self._state != TransferState.NONE:
            self._abort()

        cmd, index, subindex = SDO_STRUCT.unpack(msg[:4])

        variable = self._server.lookup(index, subindex)

        if variable.access not in ("rw", "wo"):
            raise SDODomainAbort(0x06010002)  # write a read-only object

        self._init(TransferState.BLOCK)
        self._variable = variable

        self._crc = 0 if cmd & 0x04 else None

        if cmd & 0x02:  # size bit
            size = struct.unpack("<I", msg[4:])
        else:
            size = None

        if self._handler_callback:
            self._handler = self._handler_callback(self._server.node, variable, size)

        self._sequence_number = 1

        cmd = 0xA4
        self._server.node.adapter.send(
            self._server.cob_tx,
            cmd.to_bytes(1, "little") + msg[1:4] + b"\x7F" + bytes(3),
        )

    def download_sub_block(self, msg: bytes):
        if self._sequence_number != msg[0] & 0x7F:
            self._abort()
            raise SDODomainAbort(0x05040003, self._variable)  # invalid sequence number

        last_sub_block = bool(msg[0] & 0x80)

        data = msg[1:]

        if self._crc is not None and not last_sub_block:
            self._crc = crc_hqx(data, self._crc)

        if not last_sub_block:
            self._receive(data)
        else:
            self._buffer.extend(data)
            self._state = TransferState.BLOCK_END

        if self._sequence_number == 127 or last_sub_block:
            self._server.node.adapter.send(
                self._server.cob_tx,
                b"\xA2"
                + self._sequence_number.to_bytes(1, "little")
                + b"\x7F"
                + bytes(5),
            )
            self._sequence_number = 1
        else:
            self._sequence_number += 1

    def download_block_end(self, msg: bytes):
        if self._state != TransferState.BLOCK_END:
            self._abort()
            raise SDODomainAbort(
                0x05040001, variable=False
            )  # client command specificer not valid

        size = 7 - ((msg[0] >> 2) & 0x07)

        if self._crc is not None:
            self._crc = crc_hqx(self._buffer[-7:][:size], self._crc)
            if self._crc != struct.unpack("<H", msg[1:3])[0]:
                self._abort()
                raise SDODomainAbort(0x05040004, self._variable)  # CRC invalid

        if self._handler:
            self._receive(self._buffer[:size])
        elif size != 7:
            del self._buffer[size % 7 - 7 :]

        self._finish()

        self._server.node.adapter.send(self._server.cob_tx, b"\xA1" + bytes(7))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

""" The setup script."""

from setuptools import find_packages, setup


with open('README.rst', 'rb') as readme_file:
    readme = readme_file.read().decode('utf-8')


setup(
    name='durand',
    use_scm_version={"write_to": "durand/_version.py"},
    description='CANopen library providing functionality to implement nodes',
    long_description=readme,
    long_description_content_type='text/x-rst',
    url='https://github.com/semiversus/python-durand',
    author='Günther Jena',
    author_email='guenther@jena.at',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    python_requires='>=3.7',
    license='MIT license',
    keywords='canopen can node ds301',
    packages=find_packages(include=['durand*']),
    zip_safe=False,
)



import struct
from enum import Enum
from typing import TYPE_CHECKING
import logging

from durand.datatypes import DatatypeEnum as DT
from durand.object_dictionary import Variable


if TYPE_CHECKING:
    from durand.node import Node


log = logging.getLogger(__name__)


SDO_STRUCT = struct.Struct("<BHB")


class SDODomainAbort(Exception):
    def __init__(self, code: int, variable: Variable = None):
        Exception.__init__(self)
        self.code = code
        self.variable = variable


TransferState = Enum("TransferState", "NONE SEGMENT BLOCK BLOCK_END")


class SDOServer:
    def __init__(self, node: "Node", index=0, cob_rx: int = None, cob_tx: int = None):
        self._node = node

        from .download import DownloadManager
        from .upload import UploadManager

        self.download_manager = DownloadManager(self)
        self.upload_manager = UploadManager(self)

        if cob_rx is None:
            self._cob_rx = 0x80000000
        else:
            self._cob_rx = cob_rx

        if cob_tx is None:
            self._cob_tx = 0x80000000
        else:
            self._cob_tx = cob_tx

        od = self._node.object_dictionary
        od.add_object(
            Variable(0x1200 + index, 0, DT.UNSIGNED8, "const", 3 if index else 2)
        )

        cob_rx_var = Variable(
            0x1200 + index, 1, DT.UNSIGNED32, "rw" if index else "ro", self._cob_rx
        )
        od.add_object(cob_rx_var)

        cob_tx_var = Variable(
            0x1200 + index, 2, DT.UNSIGNED32, "rw" if index else "ro", self._cob_tx
        )
        od.add_object(cob_tx_var)

        if index:
            od.download_callbacks[cob_rx].add(self._update_cob_rx)
            od.download_callbacks[cob_tx].add(self._update_cob_tx)

            od.add_object(Variable(0x1200 + index, 3, DT.UNSIGNED8, "rw"))

        self._node.adapter.add_subscription(self._cob_rx, self.handle_msg)

    @property
    def node(self):
        return self._node

    def _update_cob_rx(self, value: int):
        # bit 31: 0 - valid, 1 - invalid
        if ((self._cob_rx & self._cob_tx) & (1 << 31)) and not value & (1 << 31):
            self._node.remove_subscription(self._cob_rx & 0x7FF)

        if (
            not self._cob_rx & (1 << 31)
            and value & (1 << 31)
            and self._cob_tx & (1 << 31)
        ):
            self._node.adapter.add_subscription(value & 0x7FF)

        self._cob_rx = value

    def _update_cob_tx(self, value: int):
        # bit 31: 0 - valid, 1 - invalid
        if ((self._cob_rx & self._cob_tx) & (1 << 31)) and not value & (1 << 31):
            self._node.remove_subscription(self._cob_rx & 0x7FF)

        if (
            not self._cob_tx & (1 << 31)
            and value & (1 << 31)
            and self._cob_rx & (1 << 31)
        ):
            self._node.adapter.add_subscription(self._cob_rx & 0x7FF)

        self._cob_rx = value

    @property
    def cob_tx(self):
        if self._cob_tx & (1 << 31):
            return None

        return self._cob_tx

    @cob_tx.setter
    def cob_tx(self, cob: int):
        if cob is None:
            self._update_cob_tx(1 << 31)
        else:
            self._update_cob_tx(cob)

    @property
    def cob_rx(self):
        if self._cob_rx & (1 << 31):
            return None

        return self._cob_rx

    @cob_rx.setter
    def cob_rx(self, cob: int):
        if cob is None:
            self._update_cob_rx(1 << 31)
        else:
            self._update_cob_rx(cob)

    def handle_msg(self, cob_id: int, msg: bytes):
        assert (
            cob_id == self._cob_rx
        ), "Cob RX id invalid (0x{cob_id:X}, expected 0x{self._cob_rx:X})"

        try:
            ccs = (msg[0] & 0xE0) >> 5

            if msg[0] == 0x80:  # abort
                self.abort(msg)
            elif self.download_manager.block_transfer_active:
                self.download_manager.download_sub_block(msg)
            elif self.upload_manager.block_transfer_active:
                self.upload_manager.upload_sub_block(msg)
            elif ccs == 0:  # download segments
                self.download_manager.download_segment(msg)
            elif ccs == 1:  # init segmented download
                self.download_manager.init_download(msg)
            elif ccs == 2:  # init block upload
                self.upload_manager.init_upload(msg)
            elif ccs == 3:  # init segmented upload
                self.upload_manager.upload_segment(msg)
            elif ccs == 5 and msg[0] & 0x03 == 0:
                self.upload_manager.init_upload(msg)
            elif ccs == 6:  # init block download
                if msg[0] & 0x01:  # end block transfer
                    self.download_manager.download_block_end(msg)
                else:  # init block transfer
                    self.download_manager.download_block_init(msg)
            else:
                raise SDODomainAbort(0x05040001)  # SDO command not implemented
        except Exception as e:
            code = 0x08000000  # general error

            _, index, subindex = SDO_STRUCT.unpack(msg[:4])

            if isinstance(e, SDODomainAbort):
                code = e.code
                if e.variable:
                    index, subindex = e.variable.multiplexor
                elif e.variable is False:
                    index, subindex = 0, 0
            else:
                log.exception("Exception during processing %r", msg)

            response = SDO_STRUCT.pack(0x80, index, subindex)
            response += struct.pack("<I", code)
            self._node.adapter.send(self._cob_tx, response)

    def lookup(self, index: int, subindex: int) -> Variable:
        try:
            return self._node.object_dictionary.lookup(index, subindex)
        except KeyError:
            try:
                self._node.object_dictionary.lookup(index, 0)
            except KeyError:
                raise SDODomainAbort(0x06020000)  # object does not exist

        raise SDODomainAbort(0x06090011)  # subindex does not exist

    def abort(self, msg_data: bytes):
        _, index, subindex = SDO_STRUCT.unpack(msg_data[:4])

        try:
            variable = self._node.object_dictionary.lookup(index, subindex)
        except KeyError:
            return

        self.download_manager.on_abort(variable)
        self.upload_manager.on_abort(variable)



""" Interfacing python-canopen-node with python-can library
"""
from typing import Dict, Callable
from threading import Lock

import can

from .base import AdapterABC


class CANAdapter(AdapterABC):
    def __init__(self, *args, loop=None, **kwargs):
        self._bus = can.Bus(*args, **kwargs)
        self._loop = loop

        self.lock = Lock()
        self.subscriptions = dict()

        listener = NodeListener(self)
        can.Notifier(self._bus, (listener,), 1, self._loop)

    def add_subscription(self, cob_id: int, callback):
        with self.lock:
            self.subscriptions[cob_id] = callback

    def remove_subscription(self, cob_id: int):
        with self.lock:
            self.subscriptions.pop(cob_id)

    def send(self, cob_id: int, msg: bytes):
        msg = can.Message(arbitration_id=cob_id, data=msg, is_extended_id=False)
        self._bus.send(msg)


class NodeListener(can.Listener):
    def __init__(self, adapter: CANAdapter):
        self._adapter = adapter

    def on_message_received(self, msg: can.Message):
        if msg.is_error_frame or msg.is_remote_frame or msg.is_fd:
            # rtr is currently not supported
            return

        with self._adapter.lock:
            callback = self._adapter.subscriptions.get(msg.arbitration_id, None)

        if not callback:
            return

        callback(msg.arbitration_id, msg.data)



""" Interfacing python-canopen-node with python-canopen
"""
from typing import Dict, Callable

from canopen import Network
from .base import AdapterABC


class CANopenAdapter(AdapterABC):
    def __init__(self, network: Network):
        self._network = network
        self._subscriptions = dict()

    def add_subscription(self, cob_id: int, callback):
        self._subscriptions[cob_id] = callback
        self._network.subscribe(cob_id, callback)

    def remove_subscription(self, cob_id: int):
        callback = self._subscriptions.pop(cob_id)
        self._network.unsubscribe(cob_id, callback)

    def send(self, cob_id: int, msg: bytes):
        self._network.notify(cob_id, msg)



from typing import TYPE_CHECKING
import struct
import logging

from durand.object_dictionary import Variable

if TYPE_CHECKING:
    from durand.node import Node


log = logging.getLogger(__name__)


class TPDO:
    def __init__(self, node: "Node", index: int):
        self._node = node
        self._cob_id = 0x180 + (index - 1) * 0x100
        self._objects = ()

    def map_objects(self, *variables: Variable):
        for variable in self._objects:
            self._node.object_dictionary.update_callbacks[variable].remove(
                self._on_change
            )

        for variable in variables:
            self._node.object_dictionary.update_callbacks[variable].add(self._on_change)

        self._objects = variables

    def _on_change(self, _value):
        self._transmit()

    def _transmit(self):
        msg = b""
        for variable in self._objects:
            value = self._node.object_dictionary.read(variable)

            try:
                msg += variable.pack(value)
            except struct.error:
                log.error(f"Variable {variable!r} could not pack value {value!r}")
                return
        self._node.adapter.send(self._cob_id + self._node.node_id, msg)


class RPDO:
    def __init__(self, node: "Node", index: int):
        self._node = node

        cob_id = 0x200 + (index - 1) * 0x100 + node.node_id
        node.adapter.add_subscription(cob_id=cob_id, callback=self.handle_msg)

        self._objects = ()

    def map_objects(self, *variables: Variable):
        self._objects = variables

    def handle_msg(self, cob_id: int, msg: bytes):
        for variable in self._objects:
            value = variable.unpack(msg[: variable.size])
            self._node.object_dictionary.write(variable, value)
            msg = msg[variable.size :]



import asyncio

from durand.object_dictionary import ObjectDictionary
from durand.datatypes import DatatypeEnum
from durand import Node, Variable
from durand.adapters.can import CANAdapter


adapter = CANAdapter(interface='socketcan', channel='vcan0', loop=asyncio.get_event_loop())

node = Node(adapter, 0x0E)

state_control = Variable(0x2000, 0, DatatypeEnum.UNSIGNED8, access='rw', default=0)
node.object_dictionary.add_object(state_control)

node.tpdo[1].map_objects(state_control)
node.rpdo[1].map_objects(state_control)

# change value
node.object_dictionary.write(state_control, 10)

# read value
node.object_dictionary.read(state_control)
asyncio.get_event_loop().call_later(2, node.object_dictionary.write, state_control, 11)

data_var = Variable(0x2001, 0, DatatypeEnum.DOMAIN, access='rw')
node.object_dictionary.add_object(data_var)

node.object_dictionary.update_callbacks[data_var].add(print)

class MyDownloadHandler:
    def __init__(self):
        self._file = open('update.bin', 'wb')
    
    def on_receive(self, data: bytes):
        self._file.write(data)

    def on_finish(self):
        self._file.close()

    def on_abort(self):
        self._file.close()

#define handler
def download_callback(node, variable, size):
    if variable.index == 0x2001:
        return MyDownloadHandler()
    
    return None

node.sdo_servers[0].download_manager.set_handler_callback(download_callback)

asyncio.get_event_loop().run_forever()